;; guile-lib
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;Defines a function to generate documentation for a module. The format
;;of the documentation is @code{stexi} (@pxref{(sxml-texinfo),(sxml
;;texinfo)}).
;;
;;; Code:

(define-module (scheme module-documentation)
  #:use-module (oop goops)
  #:use-module (sxml texinfo)
  #:use-module (srfi srfi-13)
  #:use-module (scheme session)
  #:use-module (ice-9 documentation)
  #:export (module-documentation))

;; List for sorting the definitions in a module
(define defs
  '(deftp defcv defivar deftypeivar defop deftypeop defmethod
    deftypemethod defopt defvr defvar deftypevr deftypevar deffn
    deftypefn defmac defspec defun deftypefun))

(define (sort-defs . args)
  (define (priority def)
    (list-index defs (car def)))
  (define (name def)
    (cadr (assq 'name (cdadr def))))
  (let ((priorities (map priority (map cadr args))))
    (cond
     ((not (apply eq? priorities)) (apply < priorities))
     (else (apply string<=? (map name (map cadr args)))))))

(define (list*-join l infix restfix)
  (let lp ((in l) (out '()))
    (cond ((null? in) (reverse! out))
          ((symbol? in) (reverse! (cons* in restfix out)))
          (else (lp (cdr in) (if (null? out)
                                 (list (car in))
                                 (cons* (car in) infix out)))))))

(define (process-args args)
  (map (lambda (x) (if (symbol? x) (symbol->string x) x))
       (list*-join (or args '())
                   " " " . ")))

(define (get-proc-args proc)
  (process-args (and=> (procedure-source proc) cadr)))

;; This is really nasty, I wish guile gave a better way to get this...
(define (get-macro-args macro)
  (process-args (false-if-exception
                 (local-eval
                  ;; f is the local name given to the procedure we want
                  '(cadr (procedure-source f))
                  (procedure-environment (macro-transformer macro))))))

(define <class>
  (module-ref (resolve-interface '(oop goops)) '<class>)) ;; HACK!

(define (texi? str)
  (and (string-index str #\@)
       (or (string-contains str "@var")
           (string-contains str "@code")
           (string-contains str "@end"))))

(define (module-documentation sym-name)
  "Return documentation for the module named @var{sym-name}. The
documentation will be formatted as @code{stexi}
(@pxref{(sxml-texinfo),(sxml texinfo)})."
  (let* ((commentary (and=> (module-commentary sym-name)
                            string-trim-both))
         (stexi (or (cond
                     ((or (not commentary) (string-null? commentary))
                      `(*fragment* (para "[undocumented]")))
                     ((texi? commentary)
                      (false-if-exception
                       (texi-fragment->stexi commentary)))
                     (else #f))
                     `(*fragment* (verbatim ,commentary))))
         (name-str (with-output-to-string
                     (lambda () (display sym-name))))
         (module (resolve-interface sym-name)))
    (define (make-defs)
      (sort!
       (module-map
        (lambda (sym var)
          (let* ((val (and (variable-bound? var) (variable-ref var)))
                 (doc (or (and=> val object-documentation)
                          (begin (warn "variable unbound" sym) "[unbound]")))
                 (anchor-name (string-join (map symbol->string
                                                (append sym-name (list sym)))
                                           "-"))
                 (stexi (cond
                         ((string? doc)
                          (or (and (texi? doc)
                                   (false-if-exception
                                    (texi-fragment->stexi doc)))
                              `(*fragment* (verbatim ,doc))))
                         ((list? doc) doc) ;; assume it's stexi
                         (else '(*fragment* (para "[undocumented]"))))))
            (define (make-def type args)
              `((anchor (% (name ,anchor-name)))
                (,type (% ,@args) ,@(cdr stexi))))
            ;; An interesting dilemma here: How to document generic
            ;; functions? Really we are just documenting methods on
            ;; classes, not operations in general. Let's just document
            ;; methods with classes. (TODO)
            (cond
             ((and (pair? (cdr stexi)) (pair? (cadr stexi))
                   (memq (caadr stexi) defs))
              `((anchor (% (name ,anchor-name)))
                ,@(cdr stexi)))
             ((is-a? val <class>)
              (make-def 'deftp `((name ,(symbol->string sym))
                                 (category "Class"))))
             ((is-a? val <macro>)
              (make-def 'defspec `((name ,(symbol->string sym))
                                   (arguments ,@(get-macro-args val)))))
             ((is-a? val <procedure>)
              (make-def 'defun `((name ,(symbol->string sym))
                                 (arguments ,@(get-proc-args val)))))
             ((is-a? val <generic>)
              (make-def 'defop `((name ,(symbol->string sym))
                                 (category "Generic"))))
             (else
              (make-def 'defvar `((name ,(symbol->string sym))))))))
        module)
       sort-defs))

    `(texinfo (% (title ,name-str))
              (node (% (name
                        ,(string-append
                          "("
                          (string-join (map symbol->string sym-name) "-")
                          ")"))))
              (section "Overview")
              ,@(cdr stexi)
              (section "Usage")
              ,@(apply append! (make-defs)))))

;;; arch-tag: bbe2bc03-e16d-4a9e-87b9-55225dc9836c
